% Copyright (c) 2008, Rob Fergus and Jon Barron (NYU)
% All rights reserved.
%
% Redistribution and use in source and binary forms, with or without
% modification, are permitted provided that the following conditions are met:
%     * Redistributions of source code must retain the above copyright
%       notice, this list of conditions and the following disclaimer.
%     * Redistributions in binary form must reproduce the above copyright
%       notice, this list of conditions and the following disclaimer in the
%       documentation and/or other materials provided with the distribution.
%     * Neither the name of NYU nor the
%       names of its contributors may be used to endorse or promote products
%       derived from this software without specific prior written permission.
%
% THIS SOFTWARE IS PROVIDED BY ROB FERGUS AND JON BARRON ``AS IS'' AND ANY
% EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
% WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
% DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
% DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
% (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
% LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
% ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
% (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
% SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

function [f,df,f_pull,f_push, f_decay] =  CG_DrLIM(weights_vec, dimensions, data, neighbors, hidbias_offset, hidlinear, vislinear, drlim_params)

  % LINEAR HIDDEN UNITS HAVE NO MEANING IN CROSS-ENTROPY
  assert(strcmp(drlim_params.distance, 'euclidian') || ~hidlinear);

  numdata = size(data,1);
  numlayers = size(dimensions,1);

  %%% decant vector of weights into matrices
  for i=1:numlayers

    if (i==1)
      prev_dims=0;
    else
      prev_dims = sum(prod(dimensions(1:i-1,:),2));
    end

    w{i} = weights_vec(prev_dims+1:prev_dims+prod(dimensions(i,:)));
    w{i} = reshape(w{i},dimensions(i,:));
    b{i} = ones(numdata,1) * w{i}(end,:);
    w{i} = w{i}(1:end-1,:);

    dw{i} = zeros(size(w{i}));
  end

  data_orig = data;
  
  %%%% FORWARD PASS
  % run up
  pp = cell(1,numlayers);
  
  for i = 1:numlayers %%% only upto code layer
    if  (i==1) && (vislinear) % Hiddens on Gaussian visible have bias term
      pp{i} = 1./(1 + exp(-data*w{i} - b{i}));
    elseif (i==numlayers) && (hidlinear) % Gaussian top level
      pp{i} = pp{i-1}*w{i}+b{i};
    else
      pp{i} = 1./(1 + exp(-pp{i-1}*w{i} - b{i}));
    end
  end

  if i == 1
    layer_before = data;
  else
    layer_before = pp{i-1};
  end
  
  if hidlinear
    ppi_noisy = (-layer_before*w{i} - b{i} + drlim_params.noise_inject);
  else
    ppi_noisy = (1./(1 + exp(-layer_before*w{i} - b{i} + drlim_params.noise_inject)));
  end
  probs_noisy = ppi_noisy';

  % Inject noise to one half of the network
  if drlim_params.randomize_signal
    pp{i}(2:2:end,:) = ppi_noisy(2:2:end,:);
  end

  probs = pp{i}';

  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  %%%% Start of DrLIM

  % dimensions
  codelen = size(probs,1);
  numdata = size(probs,2);

  if ~hidlinear
    prob_margin = (1e-9); % Where we cutoff probs, to prevent Inf/Nan situations

    probs(probs > (1-prob_margin)) = (1-prob_margin);
    probs(probs < (prob_margin)) = (prob_margin);
  end
  
  if strcmp(drlim_params.distance, 'cross-entropy') || ...
      strcmp(drlim_params.distance, 'cross-entropy-squared')
    CE3 = zeros(codelen,numdata);
    dCE3 = zeros(codelen,numdata);
    
    % Precalc some stuff
    log_2 = log(2);
    probs_inv = 1-probs;
    log2_probs = log(probs)/log_2;
    log2_probs_inv = log(probs_inv)/log_2;
    
  elseif strcmp(drlim_params.distance, 'euclidian')
    dist3 = zeros(codelen,numdata);
  end
  
  bitDiff = zeros(codelen,numdata);
  
  % 'd' denotes the direction of the cross-entropy
  for d = 1:2
    % idx1 and idx2 are the indices of relative pairs.
    idx1 = d:2:size(probs,2);
    idx2 = (3-d):2:size(probs,2);

    if strcmp(drlim_params.distance, 'cross-entropy') || ...
        strcmp(drlim_params.distance, 'cross-entropy-squared')
      % compute cross-entropy and d/dw(cross-entropy) btw. pairs
      
      CE3(:,idx1) = -( ...
				(probs(:,idx1) .* (log2_probs(:,idx2) - log2_probs(:,idx1))) ... 
				+ (probs_inv(:,idx1) .* (log2_probs_inv(:,idx2) - log2_probs_inv(:,idx1))) ...
				);
				
      dCE3(:,idx1) = -( ...
				((log2_probs(:,idx1) +  log2_probs_inv(:,idx2)) ...
				- (log2_probs(:,idx2) + log2_probs_inv(:,idx1))) ...
				+ 1/log_2*(probs(:,idx1) - probs(:,idx2)) ...
				./(probs(:,idx1) .* probs_inv(:,idx1)) ...
				);
      
    elseif strcmp(drlim_params.distance, 'euclidian')
      % compute distance between pairs
      
      dist3(:,idx1) = probs(:,idx1) - (probs(:,idx2));
    end
    
    % Compute the distance between bit-strings
    bitDiff(:,idx1) = (probs(:,1:2:end)> ((1-hidlinear)*0.5)) ...
											~= (probs_noisy(:,2:2:end)>((1-hidlinear)*0.5));
    
  end

  if strcmp(drlim_params.distance, 'euclidian')
    dist3Sq = dist3.^2;
  end
  
  %   CE = sum(CE3,1)';

  M = drlim_params.m;
  N = drlim_params.n; % not really used now
  mult_pull = drlim_params.mult_pull;
  mult_push = drlim_params.mult_push;

  % Y == 1 if dissimilar.
  Y = ~reshape(repmat(neighbors,1,2)', [],1);

  if drlim_params.selective_push || drlim_params.selective_pull
    global bitDiff_start
    if isempty(bitDiff_start)
      bitDiff_start = bitDiff;
    else
      bitDiff = bitDiff_start;
    end
  end

  if drlim_params.selective_pull
    toPull = repmat(~Y, [1,codelen]) & bitDiff';
  else
    toPull = repmat(~Y, [1,codelen]);
  end
  
  if drlim_params.selective_push
    toPush = repmat( Y, [1,codelen]) & ~bitDiff';
  else
    toPush = repmat(Y, [1,codelen]);
  end
  
  if strcmp(drlim_params.distance, 'euclidian')
    
    if N > 0
      dist1 = sqrt(sum(dist3Sq,1));
      loss_pull = mult_pull * 0.5 * ~Y .* (max(0,(dist1-N)).^2)';
      df_drlim_pull = -2 * mult_pull * dist3' .* repmat(~Y .* (max(0,dist1-N)./(dist1+eps))', 1, codelen);
    else
      
      if drlim_params.robust_attraction
        sigma = drlim_params.robust_attraction_sigma;
%         sigma = 35; %BAD
        errSq = sum(toPull .* dist3Sq',2);
%         errSq = errSq(1:800); % BAD
%         loss_pull = mult_pull * 0.5 * errSq;
        loss_pull = mult_pull * 0.5 * (sigma^2) * (errSq)./(sigma^2 + errSq);
%         figure(1);
%         plot(loss_pull, 'b-'); hold on
%         plot(loss_pull_robust, 'r-');

%         df_drlim_pull = mult_pull * 0.5 * toPull .* (sigma^4 .* dist3')./(repmat(sum(dist3Sq',2), [1, codelen]) + sigma^2).^2;
        df_drlim_pull = -2 * mult_pull * toPull .* (sigma^4 .* dist3')./(repmat(errSq, [1, codelen]) + sigma^2).^2;
        
%         keyboard
        
      else
        loss_pull = mult_pull * 0.5 * sum(toPull .* dist3Sq',2);
        df_drlim_pull = -2*mult_pull * toPull .*  dist3';
      end
    end
    
    if drlim_params.push_bitwise    
      loss_push = sum(mult_push * 0.5 * toPush .* (max(0,M-abs(dist3)).^2)',2);
      df_drlim_push = 2 * mult_push * toPush .* (max(0,(M-abs(dist3))) .* sign(dist3))';
    else
      dist1 = sqrt(sum(dist3Sq,1));
      
      loss_push = mult_push * 0.5 * Y .* (max(0,M-dist1).^2)';
      df_drlim_push = 2 * mult_push * dist3' .* repmat(Y .* (max(0,M-dist1)./(dist1+eps))', 1, codelen);
      
    end
    
%     if isinf(M)
%       loss_push = sum(mult_push * 0.5 * (toPush .* -dist3Sq'),2);
%       df_drlim_push = -2 * mult_push * toPush .* -dist3';
%     elseif M == 0
%       
%       global M_mat
%       if isempty(M_mat)
% %         fprintf('constructing new M_mat\n');
%         M_mat = abs(probs);
% 
% %         M_mat(:,1:2:end) = max(M_mat(:,1:2:end), M_mat(:,2:2:end));
%         M_mat(:,1:2:end) = sqrt(M_mat(:,1:2:end).^2 + M_mat(:,2:2:end).^2);
%         M_mat(:,2:2:end) = M_mat(:,1:2:end);
%         
%         M_mat = 1.5*M_mat;
%       end    
%     else
%       
%       loss_push = sum(mult_push * 0.5 * toPush .* (max(0,M-abs(dist3)).^2)',2);
%       df_drlim_push = 2 * mult_push * toPush .* (max(0,(M-abs(dist3))) .* sign(dist3))';
%       
%     end
    
  elseif strcmp(drlim_params.distance, 'cross-entropy-squared')
    % Bitwise pull
    loss_pull = sum(mult_push * toPull .* (CE3.^2)',2);
    df_drlim_pull = mult_pull * toPull .* 2 .* (CE3.*dCE3)';
    
    % Total pull
%     loss_pull = mult_push* (~Y) .* sum(CE3.^2,1);
%     df_drlim_pull = -mult_push * (repmat(~Y, [1, codelen]) .* 2.*CE3.*dCE3');

    % bitwise push
    assert(isinf(M))
    loss_push = sum(mult_push * toPush .* -(CE3.^2)',2);
    df_drlim_push = mult_push * toPush .* -2 .* (CE3 .* dCE3)';
    
    % Total Push
%     loss_push = mult_push * (Y) .* sum(-(CE3.^2),1);
%     df_drlim_push = -mult_push * (repmat( Y, [1, codelen]) .* dCE3');
    
  elseif strcmp(drlim_params.distance, 'cross-entropy')
    % There is only one kind of pull (bitwise == total)
    loss_pull = mult_pull*(~Y).*sum(CE3,1)';
    df_drlim_pull = mult_pull * toPull .* dCE3';

    % Bitwise Push
    if isinf(M)
      loss_push = sum(mult_push * toPush .* -CE3',2);
    else
      loss_push = sum(mult_push * toPush .* max(0, M-CE3)',2);
    end
    df_drlim_push = -mult_push * (toPush & (CE3<M)') .* dCE3';
    
    % Total Push
    % loss_push = mult_push* (Y).*max(0,M-CE);
    % df_drlim_push = -mult_push * (repmat( Y & (CE<M), [1, codelen]) .* dCE3');
  end  
  
  %    loss  = loss_pull + loss_push;
  df_drlim = -(df_drlim_pull + df_drlim_push)';

  f_pull = sum(loss_pull(:));
  f_push = sum(loss_push(:));
  f = f_pull + f_push;

  % Decay term on total "probs"
  if (drlim_params.decay>0)
    loss_decay = drlim_params.decay * (probs.^2)';
    df_decay = drlim_params.decay * 2*(probs);
    f_decay = sum(loss_decay(:));
    
    df_drlim = df_drlim + df_decay;
    f = f + f_decay;
  else
    f_decay = 0;
  end
  
  % End of DrLIM
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  %%%%%%%%%%%%%%
  % Now backprop
  %%%% BACKWARD PASS

  t = cell(1,numlayers);
  
  for i = numlayers:-1:1

    if  (i==1) && (i~=numlayers)
      t{i} =   pp{i} .* (1-pp{i}) .* (t{i+1}*w{i+1}');
      dw{i} = [data_orig'*t{i};sum(t{i},1)];

    else

      if (i==numlayers) && hidlinear
        t{i} = df_drlim';
      elseif (i==numlayers) && (hidlinear~=1)
        t{i} = pp{i} .* (1-pp{i}) .* df_drlim';
      else
        t{i} = pp{i} .* (1-pp{i}) .* (t{i+1}*w{i+1}');
      end

      if i == 1
        dw{i} = [data_orig'*t{i};sum(t{i},1)];
      else
        dw{i} = [pp{i-1}'*t{i};sum(t{i},1)];
      end

    end
  end
  
  if drlim_params.only_train_top
    for i = 1:(numlayers-1)
      dw{i} = 0 * dw{i};
    end
  end

  %%% Decant weights back into vector
  df = [];
  for i=1:numlayers
    df = [ df ; dw{i}(:) ];
  end

  df = [df];
